/*
 *      #####################################################################
 *
 *        Copyright (C) 2008  Johan H. Bodin SM6LKM
 *
 *        This software is provided 'as is', without warranty of any kind,
 *        express or implied. In no event shall the author be held liable
 *        for any damages arising from the use of this software.
 *
 *        Permission to use, copy, modify, and distribute this software and
 *        its documentation for non-commercial purposes is hereby granted,
 *        provided that the above copyright notice and this disclaimer appear
 *        in all copies and supporting documentation.
 *
 *        The software must NOT be sold or used as part of any commercial
 *        or "non-free" product.
 *
 *      #####################################################################
 */


/*******************************************************************************

        Project:
          MEPT_JT Beacon Controller for AD9850 DDS and/or VCXO exciter.

        File:
          mjb###.c

        Author:
          Johan Bodin SM6LKM

        Processor:
          PIC16F628

        Clock:
           3.6864MHz crystal in XT mode

        Compiler:
          BKD CC5X C v3.3 extended edition

        Dependencies:
          16c628.h (supplied with compiler)
          int16cxx.h (supplied with compiler)

        Known bugs or limitations:
          None.

        - Revision history -
        ==============================================================
        Version  Date    Comment / Programmer
        --------------------------------------------------------------
        0.00     080313  Created /JHB
        0.01     080331  Big bang test, seems to work OK /JHB
        0.02     080402  - Added programmable random timeslot TX.
                         - Enabled PORTB internal weak pull-up.
                         - Added external UTC sync on RB3 --_--.
                         - Turned off the LEDs after WRITE command.
                         - Added new commands for VCXO adjustment.
                         - Added MEPT_JT baudrate calibration.
                         - Some minor fixes. /JHB
        ==============================================================

*******************************************************************************/

#include <16f628.h>
#include <int16cxx.h>


/*------------------------------------------------------------------------------
 *
 *      Set PIC configuration word
 *
 *      Cc  = CP1:CP0, code protection
 *              11 = code protection OFF
 *              10 = 0x400-0x7FF protected (upper half)
 *              01 = 0x200-0x7FF protected (upper 3/4)
 *              00 = 0x000-0x7FF protected (entire code space)
 *
 *      E   = CPD, data EEPROM protection
 *              1 = unprotected
 *              0 = protected
 *
 *      L   = LVP, low voltage programming
 *              1 = enabled, RB4 has PGM function
 *              0 = disabled, RB4 is digital I/O
 *
 *      B   = BODEN, brown-out detector
 *              1 = enabled
 *              0 = disabled
 *
 *      M   = MCLRE, RA5/MCLR pin function select
 *              1 = RA5 pin is /MCLR
 *              0 = RA5 pin is digital input, /MCLR internally tied to Vdd
 *
 *      P   = /PWRTE, power-up timer
 *              1 = disabled
 *              0 = enabled
 *
 *      W   = WDTE, watchdog timer
 *              1 = enabled
 *              0 = disabled
 *
 *      XYZ = FOSC2:FOSC1:FOSC0, oscillator selection
 *              111 = ER, CLKOUT on RA6, resistor on RA7
 *              110 = ER, I/O on RA6, resistor on RA7
 *              101 = INTRC, CLKOUT on RA6, I/O on RA7
 *              100 = INTRC,I/O on RA6, I/O on RA7
 *              011 = EC, I/O on RA6, CLKIN on RA7
 *              010 = HS, crystal/resonator on RA6 and RA7
 *              001 = XT, crystal/resonator on RA6 and RA7
 *              000 = LP, crystal/resonator on RA6 and RA7
 */

// CP off
// EP off
// LVP off
// BOD on
// MCLR off
// PWRT on
// WDT on
// OSC xt
                 // CcCc-ELBMXPWYZ
#pragma config |= 0b11111101000101


/*------------------------------------------------------------------------------
 *
 *      Some funny #defines...
 */

#define FALSE   0
#define TRUE    1


/*------------------------------------------------------------------------------
 *
 *      EEPROM map
 *
 *      PIC16F628 has 128 bytes of EEPROM memory
 *
 *      Notes:
 *        EEPROM[eaSYMBOLTABLE..eaSYMBOLTABLE+40] contains the MEPT_JT message.
 *        The two LSBits (bits 1..0) of the first byte contains the first symbol.
 *
 *        EEPROM[eaBASEPHINC..eaBASEPHINC+3] contains the DDS phase increment to
 *        use for the lowest FSK frequency (when symbol == 0). LSByte is stored
 *        first.
 *
 *        EEPROM[eaFSKOFFSETS..eaFSKOFFSETS+5] contains three 16-bit unsigned
 *        DDS phase increment offsets to use for symbols 1..3, in that order.
 *        LSByte is stored first.
 *
 *        EEPROM[eaGPSBAUDCODE] contains the GPS baudrate selector:
 *          0x00 => 1200 baud
 *          0x01 => 2400 baud
 *          0x02 => 4800 baud
 *          0x03 => 9600 baud
 *          0x04 => 19200 baud
 *
 *          All other values defaults to 9600 baud.
 *
 *        EEPROM[eaTXPROB..eaTXPROB+1] contains the "TX probability".
 *        0x0000..0xFFFF corresponds to 0..100%. LSByte is stored first.
 *
 *        EEPROM[eaBAUDCAL..eaBAUDCAL+1] contains the MEPT_JT baudrate
 *        calibration constant. Signed value. LSByte is stored first.
 */

#define eaSYMBOLTABLE   0     // 41 bytes used, next free location is 41
#define eaBASEPHINC     41    // 4 bytes used, next free location is 45
#define eaFSKOFFSETS    45    // 6 bytes used, next free location is 51
#define eaGPSBAUDCODE   51    // 1 byte used, next free location is 52
#define eaTXPROB        52    // 2 bytes used, next free location is 54
#define eaBAUDCAL       54    // 2 bytes used, next free location is 56


/*------------------------------------------------------------------------------
 *
 *      EEPROM initialization (at programming time)
 */

//#define eeBASE 0x2100         // EEPROM start word address for hex file

//#pragma cdata[eeBASE+eaDUMMY] = 0x00


/*------------------------------------------------------------------------------
 *
 *      I/O pin assignments
 *
 *      RA7 - Configured as oscillator input pin
 *      RA6 - Configured as oscillator output pin
 *      RA5 - Unused, configured as normal input with 22k pull-up (also VPP)
 *      RA4 - Output, green LED, 0 = ON
 *      RA3 - Output, red LED, 0 = ON
 *      RA2 - Output, transmitter key, 1 = transmit
 *      RA1 - Output, FSK MSBit for VCXO transmitter, 1 = +2.9297 Hz
 *      RA0 - Output, FSK LSBit for VCXO transmitter, 1 = +1.4648 Hz
 *
 *      RB7 - Output, /SDATA to AD9850 DDS (also ICSP data I/O)
 *      RB6 - Output, /SCLK to AD9850 DDS (also ICSP clock input)
 *      RB5 - Output, /FQ_UD strobe to AD9850 DDS
 *      RB4 - Output, /RESET to AD9850 DDS
 *      RB3 - Input, external sync, active low
 *      RB2 - Output, UART TX
 *      RB1 - Input, UART RX
 *      RB0 - Unused, configured as output, driven low
 *
 *      Note:
 *        The DDS control signals on RB7..RB4 are INVERTED because
 *        there are inverting buffers on SM6LKM's DDS board!
 */

// Pin access macros

#define oGLEDn        PORTA.4
#define oRLEDn        PORTA.3
#define oTX_KEY       PORTA.2

#define oDDS_SDATAn   PORTB.7
#define oDDS_SCLKn    PORTB.6
#define oDDS_STROBEn  PORTB.5
#define oDDS_RESETn   PORTB.4
#define oEXT_SYNCn    PORTB.3

#define LEDSOFF_KEYOFF_FSKZERO  PORTA = 0b00011000;

// VCXO output stuff

#define VCXOPORT      PORTA
#define bnVCXO_D1     1           // bit numbers in VCXOPORT
#define bnVCXO_D0     0
#define bmVCXOMASK    0b00000011  // bit mask for VCXOPORT

//  Initial output pin levels

#define PORTA_INIT  0b00011000  // LEDs off
#define PORTB_INIT  0b11110000  // all DDS signals low (after the inverters...)

//  Initial pin directions, 0 = OUTPUT

//    Note:
//      There's a quirk in the PIC16F627/628, TRISB.1 and TRISB.2 *BOTH*
//      have to be set (input mode) for the UART I/O to work correctly.
//      In other words, the TX pin has to be configured as input as well
//      as the RX pin...

#define TRISA_INIT  0b11100000
#define TRISB_INIT  0b00001110


/*------------------------------------------------------------------------------
 *
 *      Global variables
 *
 *      PIC16F628 has 224 bytes of RAM in three banks:
 *
 *        Bank 0: 0x20-0x7F, 96 bytes including 16 bytes in the common area
 *        Bank 1: 0xA0-0xEF, 80 bytes (plus 16 bytes in the common area)
 *        Bank 2: 0x120-0x14F, 48 bytes (plus 16 bytes in the common area)
 *        Bank 3: Nothing (plus 16 bytes in the common area)
 *
 *        Common area: 0x70-0x7F/0xF0-0xFF/0x170-0x17F/0x1F0-0x1FF, 16 bytes
 */

bank0 struct
{
  char bNextVcxoSymbol;
  char bTickCycleCount;
  bit qNewFreqInQueue;
  bit qPleaseQRT;
} jIsrGlobals;

bank0 uns16 wTick10Length;


/*------------------------------------------------------------------------------
 *
 *      Interrupt Service Routine
 *
 *      This is a tick ISR driven by CCP1 compare interrupt
 *
 *      Tsymbol = 8192/12000 = 0.68266666... s
 *      CPU Tcy = 4/3.6864e6 = 1,08506944444...e-6 s
 *      Tcy count per symbol = (8192/12000) / (4/3.6864e6) = 629145.6
 *      TMR1 is running without prescaler => full TMR1 cycle = 65536 Tcy
 *      Tsymbol happens to be exactly 9.6 full TMR1 cycles.
 *      0.6 cycles = 0.6 * 65536 = 39321.6 Tcy which rounds to 39322.
 *
 *      With calibration constant set to zero, the optimum Fxtal = 4/Tcy =
 *      4 / ((8192/12000) / 629146) = 3.68640234375 MHz.
 *      The purist may wish to pull the rock +2.3 Hz... ;-D
 *
 *      If the clock is too much off target, and you don't want to pull
 *      the crystal "physically", you can use the software calibration:
 *
 *        CAL_VALUE (command W) = 64*Fxtal/375 - 629146
 *
 *      Note:
 *        The main loop loads global wTick10Length with a calibrated Tcy
 *        count (39322 +/- calibration) just before starting transmission.
 */

#pragma origin 4

interrupt vIntServer (void)
{
  int_save_registers                        // save W and STATUS

  if (--jIsrGlobals.bTickCycleCount == 0)   // 10:th tick = time for symbol transition?
  {
    // Update outputs *first* to avoid time jitter due to non-isochronous
    // code below. The main loop has already loaded a new frequency into
    // the DDS chip and a new VCXO FSK symbol into jIsrGlobals.bNextVcxoSymbol.

    VCXOPORT = jIsrGlobals.bNextVcxoSymbol | (VCXOPORT & ~bmVCXOMASK);  // VCXO QSY!
    oTX_KEY = 1;                            // make sure VCXO transmitter is ON

    oDDS_STROBEn = 0;                       // DDS QSY!
    oDDS_STROBEn = 1;

    jIsrGlobals.qNewFreqInQueue = FALSE;    // it is now OK to load new a symbol

    if (jIsrGlobals.qPleaseQRT)             // was this the end of the last symbol?
    {
      oTX_KEY = 0;                          // yes, turn off VCXO transmitter
      CCP1IE = 0;                           // no more interrupts please
      oRLEDn = 1;                           // red LED off
    }
    else                                    // there is more to send so...
      oRLEDn = !oRLEDn;                     // ...toggle red LED at each symbol start

    // Every 10:th tick cycle should be shorter than a full TMR1 count

    CCPR1L += wTick10Length.low8;           // schedule next tick
    if (Carry)                              // propagate carry to MSByte
      CCPR1H++;
    CCPR1H += wTick10Length.high8;

    CCP1IF = 0;                             // clear interrupt flag
    jIsrGlobals.bTickCycleCount = 10;       // 10 ticks per baud (well, 9.6...)
  }
/********
  else
  {
    Adding 65536 to CCPR1H:CCPR1L is stupid as it won't change a 16-bit value ;-)
  }
********/

  CCP1IF = 0;                 // clear the CCP1 interrupt flag

  int_restore_registers       // restore W and STATUS
}


/*------------------------------------------------------------------------------
 *
 *      Read on-chip EEPROM
 *
 *      Argument:
 *        W = EEPROM address, 0x00..0x7F
 *
 *      Return:
 *        EEPROM data read from address bAddr
 */

char bReadEE (char W)
{
  EEADR = W;          // load EEPROM address
  RD = 1;             // read EEPROM
  return EEDATA;      // return read data
}


/*------------------------------------------------------------------------------
 *
 *      Write on-chip EEPROM
 *
 *      Argument:
 *        bAddr = EEPROM address, 0x00..0x7F
 *        bData = Data to be written
 *
 *      Return:
 *        TRUE if written data verifies OK, else FALSE
 *
 *      Note:
 *        PEIE related interrupts will be disabled for 6 Tcy!
 *        (never mind... the interrupt is not running when writing to EEPROM anyway...)
 */

bit qWriteEE (char bAddr, char bData)
{
  EEDATA = bData;     // load EEPROM data byte
  EEADR = bAddr;      // load EEPROM address
  WREN = 1;           // enable EEPROM write
  PEIE = 0;           // magic unlock sequence must be atomic
  EECON2 = 0x55;      // unlock
  EECON2 = 0xAA;
  WR = 1;             // start write operation
  PEIE = 1;           // allow interrupts again
  WREN = 0;           // disable EEPROM write
  do                  // wait for EEPROM "write complete"
    clrwdt ();
  while (WR);
  RD = 1;             // read back written data for verification
  return (EEDATA == bData); // return verification result
}


/*------------------------------------------------------------------------------
 *
 *      Serial output
 */

void vPut (char W)
{
  while (!TXIF)
    clrwdt ();
  TXREG = W;
}


void vPrint (const char *p)
{
  while (*p)
    vPut (*p++);
}


void vPutHexByte (char bData)
{
  char b;

  b = bData >> 4;
  if (b > 9)
    b += 55;
  else
    b += 48;
  vPut (b);

  b = bData & 0x0F;
  if (b > 9)
    b += 55;
  else
    b += 48;
  vPut (b);
}


/*------------------------------------------------------------------------------
 *
 *      Read a symbol from EEPROM
 *
 *      Argument:
 *        bSymbolIndex = sequential symbol number, 0..161
 *
 *      Return value:
 *        FSK value 0..3 (0 corresponds to the lowest transmitted frequency)
 */

char bGetSymbol (char bSymbolIndex)
{
  char bBitPairPos;
  char bSymbol @ bSymbolIndex;        // reuse a variable but change name - may save a RAM byte ;-)

  bBitPairPos = bSymbolIndex & 0x03;  // symbol position within EEPROM byte
  bSymbolIndex /= 4;                  // byte offset in EEPROM table
  bSymbolIndex += eaSYMBOLTABLE;      // add base address of symbol table
  bSymbol = bReadEE (bSymbolIndex);   // read 1 byte = 4 symbols from EEPROM
  bBitPairPos *= 2;                   // two bits per bit pair...
  bSymbol >>= bBitPairPos;            // shift the wanted symbol into bits 1..0
  return bSymbol & 0x03;              // mask off unwanted stuff
}


/*------------------------------------------------------------------------------
 *
 *      Dump all EEPROM settings on the serial port
 */

void vDumpSettings (void)
{
  char bTmp;
  char bSymIx;

  bTmp = 60;
  do
    vPut ('=');
  while (--bTmp);

  vPrint ("\r\n\r\nFrequencies:\r\n  ");
  bTmp = 34;
  do
    vPut ('-');
  while (--bTmp);
  vPrint ("\r\n  Symbol 0 DDS PhInc (P): 0x");
  vPutHexByte (bReadEE (eaBASEPHINC+3));  // show MSByte first
  vPutHexByte (bReadEE (eaBASEPHINC+2));
  vPutHexByte (bReadEE (eaBASEPHINC+1));
  vPutHexByte (bReadEE (eaBASEPHINC));
  vPrint ("\r\n  Symbol 1 DDS Offset (Q): +0x");
  vPutHexByte (bReadEE (eaFSKOFFSETS+1)); // show MSByte first
  vPutHexByte (bReadEE (eaFSKOFFSETS));
  vPrint ("\r\n  Symbol 2 DDS Offset (R): +0x");
  vPutHexByte (bReadEE (eaFSKOFFSETS+3));
  vPutHexByte (bReadEE (eaFSKOFFSETS+2));
  vPrint ("\r\n  Symbol 3 DDS Offset (S): +0x");
  vPutHexByte (bReadEE (eaFSKOFFSETS+5));
  vPutHexByte (bReadEE (eaFSKOFFSETS+4));
  vPrint ("\r\n  ");
  bTmp = 34;
  do
    vPut ('-');
  while (--bTmp);
  vPrint ("\r\n\r\nMessage (T):\r\n  ");
  bTmp = 49;
  do
    vPut ('-');
  while (--bTmp);
  vPrint ("\r\n  ");
  bSymIx = 0;
  bTmp = 0;
  for (;;)
  {
    vPut (bGetSymbol (bSymIx) + '0');
    bSymIx++;
    if (bSymIx == 162)
      break;
    bTmp++;
    if (bTmp == 40)
    {
      bTmp = 0;
      vPrint ("\r\n  ");
    }
    else
    if ((bTmp & 0x03) == 0)
      vPut (32);
  }
  vPrint ("\r\n  ");
  bTmp = 49;
  do
    vPut ('-');
  while (--bTmp);
  vPrint ("\r\n\r\nTX Probability (U): 0x");
  vPutHexByte (bReadEE (eaTXPROB+1)); // show MSByte first
  vPutHexByte (bReadEE (eaTXPROB));
  vPrint (" out of 0xFFFF");
  vPrint ("\r\nGPS Baud Rate Code (V): 0x");
  bTmp = bReadEE (eaGPSBAUDCODE);
  vPutHexByte (bTmp);
  vPut (32);
  vPut ('(');
  switch (bTmp)
  {
    case 0: vPrint ("1200");  break;
    case 1: vPrint ("2400");  break;
    case 2: vPrint ("4800");  break;
    default:
    case 3: vPrint ("9600");  break;
    case 4: vPrint ("19200"); break;
  }
  vPrint (")\r\nBaud Cal (W): 0x");
  vPutHexByte (bReadEE (eaBAUDCAL+1)); // show MSByte first
  vPutHexByte (bReadEE (eaBAUDCAL));
  vPrint ("\r\n\r\n");
  bTmp = 60;
  do
    vPut ('=');
  while (--bTmp);
  vPrint ("\r\n");
}


/*------------------------------------------------------------------------------
 *
 *      Programming mode
 *
 *      To unlock the programming mode, enter four '+' characters (++++)
 *      within 5 seconds after PIC reset. New data can be downloaded after
 *      the welcome message:
 *
 *      Format (ASCII):
 *        P <DDS phase increment for symbol 0 (8 hex digits)>
 *        Q <unsigned DDS phase increment offset for symbol 1 (4 hex digits)>
 *        R <unsigned DDS phase increment offset for symbol 2 (4 hex digits)>
 *        S <unsigned DDS phase increment offset for symbol 3 (4 hex digits)>
 *        T <162 message symbol digits, 0..3>
 *        U <TX probability, 0000..FFFF => 0..100% (4 hex digits)>
 *        V <GPS port baud rate code digit 0..4 (1200..19200 baud)>
 *        W <signed symbol time calibration, Tcy/symbol (4 hex digits)>
 *
 *        X = Save all entered data to EEPROM and start the beacon
 *
 *      Example:
 *        ------------------------
 *        p 03A75614 <cr><lf>
 *        q 0041 <cr><lf>
 *        r 0082 <cr><lf>
 *        s 00C3 <cr><lf>
 *        t 01023023010231023010231023012301...(162 digits) <cr><lf>
 *        u 4000 <cr><lf>
 *        v 3 <cr><lf>
 *        w FFEF (-17 => -17/921600 = -18.446us per symbol)
 *        X <cr><lf>
 *        ------------------------
 *
 *      Notes:
 *        - The download data is not case sensitive
 *        - All whitespace such as spaces, tabs, CR LF etc. is ignored
 *        - The variables may appear in any order
 *        - P, Q, R, S, T, U, V and W will terminate any entry in progress and start a new one
 *        - Excess digits are ignored
 *        - The hex data shall *not* have any prefix such as $, 0x etc.
 *
 *      Notes:
 *        - TMR0 is assumed to be running on internal clock with prescaler = 256
 */

#define tTICKSPERSECOND 14  // ~14 TMR0 overflow ticks per second

void vPrintUpdateResult (char bVarCode, bit qOkay)
{
  vPrint ("\r\n");
  vPut (bVarCode);
  if (qOkay)
    vPrint (" updated OK");
  else
    vPrint (" Failed! - EEPROM error!");
}


void vPrintNotUpdated (char bVarCode)
{
  vPrint ("\r\n");
  vPut (bVarCode);
  vPrint (" NOT updated!");
}


void vProgMode (void)
{
  char bTickDiv;
  char bSecondsToTimeout;
  char bState;
  char bRx;
  char bCmd;
  char bDigitCnt;
  char bShifter;
  uns32 ulP;
  uns16 wQ;
  uns16 wR;
  uns16 wS;
  char abT[41];
  uns16 wU;
  char bV;
  uns16 wW;
  bit qP_Updated;
  bit qQ_Updated;
  bit qR_Updated;
  bit qS_Updated;
  bit qT_Updated;
  bit qU_Updated;
  bit qV_Updated;
  bit qW_Updated;
  bit qBurnOk;

  // Initialize the UART for 9600 baud 8N1

  SPBRG = 5;                          // 9600 baud with BRGH=0 and Fxtal=3.6864MHz
  TXSTA = 0b00100000;                 // 8-bit TX, TX on, asynch, BRGH=0 (low speed)
  RCSTA = 0b10010000;                 // UART on, 8-bit RX, continous RX, address detect off

  vPrint ("\r\n\r\nMJB v002\r\n\r\nCurrent Settings:\r\n");
  vDumpSettings ();
  vPrint ("\r\nEnter ++++ for cmd mode...");

  oGLEDn = 0;                         // green LED on
  bState = 0;                         // state 0 => look for "++++" within 5 seconds
  bDigitCnt = 0;                      // no characters received yet
  bTickDiv = tTICKSPERSECOND;         // load full second
  bSecondsToTimeout = 5;              // initial timeout ~5s (for entering ++++)

  for (;;)                            // stay in this loop until "WRITE" command or any timeout
  {
    clrwdt ();                        // feed the watchdog

    // Run timeout timer

    if (T0IF)                         // TMR0 tick?
    {
      T0IF = 0;                       // yes, clear overflow flag
      if (--bTickDiv == 0)            // full second?
      {
        bTickDiv = tTICKSPERSECOND;   // yes!
        if (--bSecondsToTimeout == 0) // timeout yet?
        {
          LEDSOFF_KEYOFF_FSKZERO      // LEDs off, key = 0, FSK output = 0
//          oRLEDn = 1;                 // blow out the LEDs
//          oGLEDn = 1;
//          oTX_KEY = 0;                // turn off VCXO TX, if on
          vPrint ("\r\nStarting beacon...\r\n");
          return;                     // yes, give up, go start the beacon
        }
      }
    }

    if (!RCIF)                        // any character received by UART?
      continue;                       // no, keep on waiting, back to top of loop

    if (RCSTA & 0x06)                 // yes, overrun or framing error?
    {
      W = RCREG;                      // yes, just flush the UART's FIFO and...
      W = RCREG;
      CREN = 0;                       // ...reset the UART (ignore the error...)
      CREN = 1;
      continue;                       // keep on waiting, back to top of loop
    }

    bRx = RCREG;                      // pick the received character from the UART

    if (bState == 0)                  // still waiting for unlock string? (++++)
    {
      if (bRx == '+')                 // '+' received?
      {
        bDigitCnt++;                  // yes, count it
        if (bDigitCnt.2)              // 4 consecutive '+' received?
        {
          bSecondsToTimeout = 60;     // ~1 minute timeout after unlock
          oRLEDn = 0;                 // unlocked! red LED on
          bState = 1;                 // switch to idle state
        }
      }
      else                            // it was another character (not '+')
        bDigitCnt = 0;                // start counting from scratch
      continue;                       // go wait for the next character
    }

    // Command characters L, M, N and O just key the TX and change the FSK outputs.
    // Command characters P, Q, R, S, T, U, V and W have priority and will force
    // the simple state machine to begin receiving a new value at any time.
    // Command character X is the "save to EEPROM and exit" command.

    if ((bRx >= 'a') && (bRx <= 'z')) // convert lower case letters to upper case
      bRx -= 0x20;

    bCmd = bRx - 'L';                 // bRx = 'L'..'X' => bCmd = 0..12
    if (bCmd <= 12)                   // valid command character?
    {
      bSecondsToTimeout = 60;         // yes, valid command started, reload timeout
      bDigitCnt = 0;                  // no digits received yet for this command

      skip (bCmd);                    // jump table is faster/smaller than a switch()
      goto cmd0;                      // L => TX FSK symbol 0 (for VCXO adjustment)
      goto cmd1;                      // M => TX FSK symbol 1 (for VCXO adjustment)
      goto cmd2;                      // N => TX FSK symbol 2 (for VCXO adjustment)
      goto cmd3;                      // O => TX FSK symbol 3 (for VCXO adjustment)
      goto cmd4;                      // P => set DDS base phase increment
      goto cmd5;                      // Q => set DDS symbol 1 phase increment offset
      goto cmd6;                      // R => set DDS symbol 2 phase increment offset
      goto cmd7;                      // S => set DDS symbol 3 phase increment offset
      goto cmd8;                      // T => set beacon message
      goto cmd9;                      // U => set TX probability
      goto cmd10;                     // V => set GPS baud rate
      goto cmd11;                     // W => set MEPT_JT baud rate calibration value
      goto cmd12;                     // X => save all updated settings to EEPROM and exit

#pragma computedGoto 0

cmd0: // *#* WARNING: Ugly hardcoded access to FSK pins in PORTA!
cmd1: // bCmd == wanted symbol code and FSK outputs D1:D0 happens
cmd2: // to be on PORTA pins 1:0 so we save some code space by doing it the simple way:
cmd3:
      PORTA = (PORTA & 0b11111100) | bCmd;  // set FSK output to wanted symbol number
      oTX_KEY = 1;                    // turn on VCXO TX
      bSecondsToTimeout = 0;          // VCXO adjustment may take some time, 0 => 256 seconds timeout
      continue;                       // stay in command receiving state (1)
cmd4:
      qP_Updated = FALSE;             // flag this variable as "not updated" (yet)
      bState = 2;                     // start picking digits for P
      continue;
cmd5:
      qQ_Updated = FALSE;
      bState = 3;                     // start picking digits for Q
      continue;
cmd6:
      qR_Updated = FALSE;
      bState = 4;                     // start picking digits for R
      continue;
cmd7:
      qS_Updated = FALSE;
      bState = 5;                     // start picking digits for S
      continue;
cmd8:
      qT_Updated = FALSE;
      bState = 6;                     // start picking digits for T
      continue;
cmd9:
      qU_Updated = FALSE;
      bState = 7;                     // start picking digits for U
      continue;
cmd10:
      qV_Updated = FALSE;
      bState = 8;                     // pick a digit for V
      continue;
cmd11:
      qW_Updated = FALSE;
      bState = 9;                     // start picking digits for W
      continue;
cmd12:
      vPut (13);
      vPut (10);
      if (qP_Updated)                 // P value updated OK?
      {
        qBurnOk = qWriteEE (eaBASEPHINC, ulP.low8);   // yes, copy from buffer to EEPROM
        qBurnOk &= qWriteEE (eaBASEPHINC+1, ulP.midL8);
        qBurnOk &= qWriteEE (eaBASEPHINC+2, ulP.midH8);
        qBurnOk &= qWriteEE (eaBASEPHINC+3, ulP.high8);
        vPrintUpdateResult ('P', qBurnOk);
      }
      else
        vPrintNotUpdated ('P');

      if (qQ_Updated)
      {
        qBurnOk = qWriteEE (eaFSKOFFSETS+0, wQ.low8);
        qBurnOk &= qWriteEE (eaFSKOFFSETS+1, wQ.high8);
        vPrintUpdateResult ('Q', qBurnOk);
      }
      else
        vPrintNotUpdated ('Q');

      if (qR_Updated)
      {
        qBurnOk = qWriteEE (eaFSKOFFSETS+2, wR.low8);
        qBurnOk &= qWriteEE (eaFSKOFFSETS+3, wR.high8);
        vPrintUpdateResult ('R', qBurnOk);
      }
      else
        vPrintNotUpdated ('R');

      if (qS_Updated)
      {
        qBurnOk = qWriteEE (eaFSKOFFSETS+4, wS.low8);
        qBurnOk &= qWriteEE (eaFSKOFFSETS+5, wS.high8);
        vPrintUpdateResult ('S', qBurnOk);
      }
      else
        vPrintNotUpdated ('S');

      if (qT_Updated)
      {
        qBurnOk = TRUE;
        for (bDigitCnt = 0; bDigitCnt < sizeof(abT); bDigitCnt++)
          qBurnOk &= qWriteEE (eaSYMBOLTABLE+bDigitCnt, abT[bDigitCnt]);
        vPrintUpdateResult ('T', qBurnOk);
      }
      else
        vPrintNotUpdated ('T');

      if (qU_Updated)
      {
        qBurnOk = qWriteEE (eaTXPROB+0, wU.low8);
        qBurnOk &= qWriteEE (eaTXPROB+1, wU.high8);
        vPrintUpdateResult ('U', qBurnOk);
      }
      else
        vPrintNotUpdated ('U');

      if (qV_Updated)
      {
        qBurnOk = qWriteEE (eaGPSBAUDCODE, bV);
        vPrintUpdateResult ('V', qBurnOk);
      }
      else
        vPrintNotUpdated ('V');

      if (qW_Updated)
      {
        qBurnOk = qWriteEE (eaBAUDCAL+0, wW.low8);
        qBurnOk &= qWriteEE (eaBAUDCAL+1, wW.high8);
        vPrintUpdateResult ('W', qBurnOk);
      }
      else
        vPrintNotUpdated ('W');

      vPrint ("\r\n\r\nNew Settings:\r\n");
      vDumpSettings ();             // show new settings as stored in EEPROM
      vPrint ("\r\nStarting beacon...\r\n");
      LEDSOFF_KEYOFF_FSKZERO        // LEDs off, key = 0, FSK output = 0
//      oRLEDn = 1;                   // blow out the red LED
//      oGLEDn = 1;                   // blow out the green LED
//      oTX_KEY = 0;                  // turn off VCXO TX, if on
      return;
    }

    if (bState == 1)                  // character other than PQRSTUV in idle state?
      continue;                       // yes, simply ignore it

    // If valid hex digit, convert it to a 0..15 nibble, else ignore the character

    if ((bRx >= '0') && (bRx <= '9'))
      bRx -= '0';
    else
    if ((bRx >= 'A') && (bRx <= 'F'))
      bRx -= ('A' - 10);
    else                              // not a valid hex digit
      continue;                       // ignore it

    if (bState == 2)                  // picking 8 hex digits for P?
    {
      if ((bDigitCnt & 0b11111000) == 0)  // < 8 digits entered so far?
      {
        ulP <<= 4;                    // yes, make room for next nibble in LSByte
        ulP.low8 |= bRx;              // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.3)              // 8 digits entered? (bit 3 == 1?)
        {
          qP_Updated = TRUE;          // yes, P fully entered
          bState = 1;                 // done with P
        }
      }
      continue;
    }

    if (bState == 3)                  // picking 4 hex digits for Q?
    {
      if ((bDigitCnt & 0b11111100) == 0)  // < 4 digits entered so far?
      {
        wQ <<= 4;                     // yes, make room for next nibble in LSByte
        wQ.low8 |= bRx;               // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.2)              // 4 digits entered? (bit 2 == 1?)
        {
          qQ_Updated = TRUE;          // yes
          bState = 1;                 // done with Q
        }
      }
      continue;
    }

    if (bState == 4)                  // picking 4 hex digits for R?
    {
      if ((bDigitCnt & 0b11111100) == 0)  // < 4 digits entered so far?
      {
        wR <<= 4;                     // yes, make room for next nibble in LSByte
        wR.low8 |= bRx;               // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.2)              // 4 digits entered? (bit 2 == 1?)
        {
          qR_Updated = TRUE;          // yes
          bState = 1;                 // done with R
        }
      }
      continue;
    }

    if (bState == 5)                  // picking 4 hex digits for S?
    {
      if ((bDigitCnt & 0b11111100) == 0)  // < 4 digits entered so far?
      {
        wS <<= 4;                     // yes, make room for next nibble in LSByte
        wS.low8 |= bRx;               // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.2)              // 4 digits entered? (bit 2 == 1?)
        {
          qS_Updated = TRUE;          // yes
          bState = 1;                 // done with S
        }
      }
      continue;
    }

    if (bState == 6)                  // picking 162 0/1/2/3 digits for T?
    {
      if (bDigitCnt < 162)            // no digit entered yet?
      {
        if (bRx & 0b11111100)         // invalid symbol code? (bRx > 3?)
        {
          bState = 1;                 // yes, ignore this command
          continue;
        }

        bShifter >>= 2;               // make room for this 2-bit symbol in the shift register
        if (bRx.1)                    // put the new symbol in the two MSBits
          bShifter.7 = 1;
        if (bRx.0)
          bShifter.6 = 1;


        if (bDigitCnt == 161)         // last symbol?
        {
          bShifter = swap (bShifter); // yes, move the two last symbols into LSNibble
          bDigitCnt += 2;             // simulate "whole byte filled", 163 == 0xA3
          qT_Updated = TRUE;
          bState = 1;                 // done with T
        }

        if ((bDigitCnt & 0x03) == 3)  // whole byte filled (4 symbols) or last symbol (bDigitCnt == 0xA3)?
        {
          bRx = bDigitCnt >> 2;       // find array index (bRx used as a temp variable here)
          abT[bRx] = bShifter;        // save the 4 symbols
        }

        bDigitCnt++;                  // count the symbol
      }
      continue;
    }

    if (bState == 7)                  // picking 4 hex digits for U?
    {
      if ((bDigitCnt & 0b11111100) == 0)  // < 4 digits entered so far?
      {
        wU <<= 4;                     // yes, make room for next nibble in LSByte
        wU.low8 |= bRx;               // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.2)              // 4 digits entered? (bit 2 == 1?)
        {
          qU_Updated = TRUE;          // yes
          bState = 1;                 // done with U
        }
      }
      continue;
    }

    if (bState == 8)                  // picking 1 digit for V?
    {
      if (bDigitCnt == 0)             // no digit entered yet?
      {
        if (bRx > 4)                  // invalid baud rate code?
        {
          bState = 1;                 // yes, ignore this command
          continue;
        }
        bV = bRx;                     // code OK, save it
        bDigitCnt++;                  // count this digit
        qV_Updated = TRUE;
        bState = 1;                   // done with V
      }
      continue;
    }

    if (bState == 9)                  // picking 4 hex digits for W?
    {
      if ((bDigitCnt & 0b11111100) == 0)  // < 4 digits entered so far?
      {
        wW <<= 4;                     // yes, make room for next nibble in LSByte
        wW.low8 |= bRx;               // squeeze in the new nibble
        bDigitCnt++;                  // count this digit
        if (bDigitCnt.2)              // 4 digits entered? (bit 2 == 1?)
        {
          qW_Updated = TRUE;          // yes
          bState = 1;                 // done with W
        }
      }
      continue;
    }
  }
}


/*------------------------------------------------------------------------------
 *
 *      Load new frequency into the AD9850 DDS chip
 *      (BUT DO NOT STROBE IT YET!!!)
 *
 *      Argument:
 *        ulNewPhaseIncrement = 32-bit phase increment for DDS
 *        Fout = Fclk_dds * ulNewPhaseIncrement / 2^32
 */

void vPrepareDdsForQsyTo (uns32 ulNewPhaseIncrement)
{
  char bLoop;

  // Download 32-bit phase increment word to the DDS chip

  bLoop = 32;
  do
  {
    oDDS_SDATAn = !ulNewPhaseIncrement.0; // shift out data LSBit first (inverted!)
    oDDS_SCLKn = 0;                       // make clock pulse on AD9850 W_CLK pin
    oDDS_SCLKn = 1;
    ulNewPhaseIncrement = rr (ulNewPhaseIncrement); // shift data bits to the left
  } while (--bLoop);                      // all 32 bits transmitted?

  // Download 8 control & phase bits (not used, send 0x00)

  bLoop = 8;
  oDDS_SDATAn = 1;                        // inverted zero...
  do
  {
    oDDS_SCLKn = 0;                       // make clock pulse on AD9850 W_CLK pin
    oDDS_SCLKn = 1;
  } while (--bLoop);                      // all 8 bits transmitted?

  // The new frequency is now loaded - just pulse oDDS_STROBEn --_-- to QSY
}


/*------------------------------------------------------------------------------
 *
 *      Initialize the AD9850 DDS chip
 */

void vInitDdsChip (void)
{
  oDDS_RESETn = 0;                // reset the AD9850
  oDDS_RESETn = 1;
  oDDS_SCLKn = 0;                 // make W_CLK pulse to enter serial mode
  oDDS_SCLKn = 1;
  oDDS_STROBEn = 0;               // make FQ-UD pulse
  oDDS_STROBEn = 1;
  vPrepareDdsForQsyTo (0);        // zero Hz = DC = output OFF
  oDDS_STROBEn = 0;
  oDDS_STROBEn = 1;               // done
}


/*------------------------------------------------------------------------------
 *
 *      Prepare for next symbol transition
 *
 *      Argument:
 *        bSym = symbol, 0..3
 *
 *      Output:
 *        The two symbol bits are copied to the corresponding hardware
 *        port bit positions in jIsrGlobals.bNextVcxoSymbol. The tick
 *        ISR will then write these bits to the VCXO port at the next
 *        symbol transition.
 *
 *        The DDS phase increment corresponding to the symbol is computed
 *        and downloaded to the DDS hardware. The actual frequency shift,
 *        if any, will be occur when the tick ISR strobes the DDS chip at
 *        the next "baud tick".
 */

void vEnqueueSymbol (char bSym)
{
  uns32 ulPhInc;
  uns16 wOffset;

  // Prepare VCXO output

  jIsrGlobals.bNextVcxoSymbol = 0;              // assume symbol == 0
  if (bSym.0)                                   // copy 1's to appropriate hardware positions
    jIsrGlobals.bNextVcxoSymbol.bnVCXO_D0 = 1;
  if (bSym.1)
    jIsrGlobals.bNextVcxoSymbol.bnVCXO_D1 = 1;

  // Prepare DDS output

  ulPhInc.low8 = bReadEE (eaBASEPHINC);         // read symbol 0 phase increment from EEPROM
  ulPhInc.midL8 = bReadEE (eaBASEPHINC+1);      // this is stored LSByte first
  ulPhInc.midH8 = bReadEE (eaBASEPHINC+2);
  ulPhInc.high8 = bReadEE (eaBASEPHINC+3);

  if (bSym)                                     // symbol other than 0?
  {
    bSym--;                                     // yes, it is 1, 2 or 3, make it 0, 1 or 2
    bSym *= 2;                                  // 2 bytes in each offset table entry
    bSym += eaFSKOFFSETS;                       // add base address
    wOffset.low8 = bReadEE (bSym);              // read LSByte from EEPROM
    bSym++;                                     // bump pointer
    wOffset.high8 = bReadEE (bSym);             // read MSByte from EEPROM
    ulPhInc += wOffset;                         // now we have the phase increment for this symbol
  }

  vPrepareDdsForQsyTo (ulPhInc);                // download to DDS hardware but don't QSY yet...

  jIsrGlobals.qNewFreqInQueue = TRUE;           // this will be set FALSE by the ISR when it wants more
}


/*------------------------------------------------------------------------------
 *
 *      Synchronize to even UTC minute from GPS NMEA stream
 *
 *      This is a simple "fly-by" state machine that looks for $GPRMC and
 *      $GPGGA sentences on the serial port and returns TRUE on the top of
 *      even minutes (if the UTC time looks valid).
 *
 *      <cr-lf ignored> $GPGGA,hhmmss <rest of line ignored>
 *      <cr-lf ignored> $GPRMC,hhmmss <rest of line ignored>
 *
 *      This function must be called at a rate faster than the serial
 *      character rate.
 *
 *      The NMEA checksum, if present, is *not* checked (this is a hobby project after all ;-).
 */

bit qGpsEvenMinute (void)
{
  static char bNmeaState;           // static => cleared to zero by clearRAM() at start-up
  static char bHour;
  static char bMinute @ bHour;      // save a byte of RAM, bMinute and bHour are never used at the same time
  static char bSecond;
  char bRx;

  if (!RCIF)                        // any character received?
    return FALSE;                   // no, quit immediately

  if (RCSTA & 6)                    // UART overrun or framing error?
  {
    W = RCREG;                      // drain FIFO
    W = RCREG;
    CREN = 0;                       // reset the UART (it gets stuck after overrun error!)
    CREN = 1;
    return FALSE;                   // quit
  }

  bRx = RCREG;                      // pick the received character from the UART

  if (bRx == '$')                   // start character of NMEA sentence?
  {
    bNmeaState = 1;                 // yes, start receiving the sentence from scratch
    return FALSE;
  }

  // NMEA receiving state machine
  // "Computed Goto" using a jump table is faster than a long switch()...

  if (bNmeaState > 14)              // illegal state!? ("what if" paranoia)
    for (;;)                        // wait for watchdog reset
      GIE = 0;                      // keep interrupts disabled

  skip (bNmeaState);
  goto st0_Idle;
  goto st1_Get_Gpxxx;
  goto st2_Get_gPxxx;
  goto st3_Get_gpGga_or_gpRmc;
  goto st4_Get_gpgGa;
  goto st5_Get_gpggA;
  goto st6_Get_gprMc;
  goto st7_Get_gprmC;
  goto st8_Get_comma;
  goto st9_Get_10h;
  goto st10_Get_1h;
  goto st11_Get_10m;
  goto st12_Get_1m;
  goto st13_Get_10s;
  goto st14_Get_1s;

#pragma computedGoto 0

st0_Idle:
  return FALSE;

st1_Get_Gpxxx:
  if (bRx == 'G')               // 'G' is expected here
    bNmeaState++;               // ok, proceed to next state
  else                          // not the expected 'G' so...
    bNmeaState = 0;             // ...back to square zero and wait for another '$'
  return FALSE;

st2_Get_gPxxx:
  if (bRx == 'P')               // 'P' is expected here
    bNmeaState++;               // ok, we've got "GP" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st3_Get_gpGga_or_gpRmc:
  if (bRx == 'G')               // 'G' or 'R' is expected here
    bNmeaState++;               // ok, we've got "GPG" so far
  else
  if (bRx == 'R')
    bNmeaState = 6;             // ok, we've got "GPR" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st4_Get_gpgGa:
  if (bRx == 'G')               // 'G' is expected here
    bNmeaState++;               // ok, we've got "GPGG" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st5_Get_gpggA:
  if (bRx == 'A')               // 'A' is expected here
    bNmeaState = 8;             // ok, we've got "GPGGA" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st6_Get_gprMc:
  if (bRx == 'M')               // 'M' is expected here
    bNmeaState++;               // ok, we've got "GPRM" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st7_Get_gprmC:
  if (bRx == 'C')               // 'C' is expected here
    bNmeaState++;               // ok, we've got "GPRMC" so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st8_Get_comma:
  if (bRx == ',')               // ',' is expected here
    bNmeaState++;               // ok, we've got "GPGGA," or "GPRMC," so far
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st9_Get_10h:
  bRx -= '0';                   // hopefully the 10 hour digit, convert to binary
  if (bRx <= '2')               // 0..2?
  {
    bHour = bRx;                // yes, valid digit
    bHour <<= 3;                // multiply by 10 the fast way ;-)
    bHour += bRx;
    bHour += bRx;
    bNmeaState++;
  }
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st10_Get_1h:
  bRx -= '0';                   // hopefully the 1 hour digit, convert to binary
  if (bRx <= '9')               // 0..9?
  {
    bHour += bRx;               // yes, valid digit, add 10*10's + 1's
    if (bHour <= 23)            // valid hour?
      bNmeaState++;             // OK!
  }
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st11_Get_10m:
  bRx -= '0';                   // hopefully the 10 minute digit, convert to binary
  if (bRx <= '5')               // 0..5?
  {
    bMinute = bRx;              // yes, valid digit
    bMinute <<= 3;              // multiply by 10 the fast way ;-)
    bMinute += bRx;
    bMinute += bRx;
    bNmeaState++;
  }
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st12_Get_1m:
  bRx -= '0';                   // hopefully the 1 minute digit, convert to binary
  if (bRx <= '9')               // 0..9?
  {
    bMinute += bRx;             // yes, valid digit, add 10*10's + 1's
    if (bMinute <= 59)          // valid minute?
      bNmeaState++;             // OK!
  }
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st13_Get_10s:
  bRx -= '0';                   // hopefully the 10 second digit, convert to binary
  if (bRx <= '5')               // 0..5?
  {
    bSecond = bRx;              // yes, valid digit
    bSecond <<= 3;              // multiply by 10 the fast way ;-)
    bSecond += bRx;
    bSecond += bRx;
    bNmeaState++;
  }
  else
    bNmeaState = 0;             // sorry...
  return FALSE;

st14_Get_1s:
  bNmeaState = 0;               // back to idle (this is the last state anyway)
  bRx -= '0';                   // hopefully the 1 second digit, convert to binary
  if (bRx <= '9')               // 0..9?
  {
    bSecond += bRx;             // yes, valid digit, add 10*10's + 1's
    oGLEDn = !oGLEDn;           // valid time found, toggle green LED (0.5Hz)
    if ((bMinute.0 == 0) && (bSecond == 0)) // top of even minute?
      return TRUE;              // GO!
  }
  return FALSE;
}


/*------------------------------------------------------------------------------
 *
 *      Refresh pin modes, feed the dog, refresh OPTION, turn off analog stuff
 */

void vRefreshMCU (void)
{
  TRISA = TRISA_INIT;   // set default pin directions
  TRISB = TRISB_INIT;

  // Set up the OPTION register:
  //
  // Bit 7 = /RBPU  = 0, internal PORTB pull-up ON
  // Bit 6 = INTEDG = 1, RB0 interrupt on rising edge (not used)
  // Bit 5 = T0CS   = 0, internal TMR0 clock
  // Bit 4 = T0SE   = 1, TMR0 counts on falling edge on RA4 (not used)
  // Bit 3 = PSA    = 0, prescaler assigned to TMR0 (not WDT)
  // Bit 2..0 = PS2..PS0 = 111, TMR0 prescaler = 256 => 71.111ms overflow rate

  clrwdt ();
  OPTION = 0b01010111;

  VRCON = 0;            // Vref module OFF to save a few electrons
  CMCON = 0b00000111;   // comparators OFF, all PORTA pins digital
  INTCON = 0b11000000;  // GIE = PEIE = 1 (needed for CCP1 interrupt)
}


/*------------------------------------------------------------------------------
 *
 *      Main program - This is the "reset entry point"
 */

void main (void)
{
  static char bSymbolIndex;
  static uns16 wRand;
  static bit qPrevExtSyncSample;
  bit qExtSyncEdge;

  // Initialize the CPU

  PORTA = PORTA_INIT;   // set initial output pin levels
  PORTB = PORTB_INIT;

  vRefreshMCU ();       // set pin modes, critical registers etc.

  clearRAM ();          // clear entire RAM

  vInitDdsChip ();      // put the DDS in a known state

  vProgMode ();         // go check if user wants to download some data

  // Initialize UART for GPS NMEA reception

  RCSTA = 0;                    // reset UART

  switch (bReadEE (eaGPSBAUDCODE))
  {
    case 0: SPBRG = 47; break;  // 1200 baud with BRGH=0 and Fxtal=3.6864MHz
    case 1: SPBRG = 23; break;  // 2400 baud
    case 2: SPBRG = 11; break;  // 4800 baud
    default:
    case 3: SPBRG = 5;  break;  // 9600 baud
    case 4: SPBRG = 2;  break;  // 19200 baud
  }
  TXSTA = 0b00100000;   // 8-bit TX, TX on, asynch, BRGH=0 (low speed)
  RCSTA = 0b10010000;   // UART on, 8-bit RX, continous RX, address detect off

  // Here is the main "eternal" loop

  for (;;)
  {
    // Keep pins pointing in appropriate directions (may change due to ESD) etc.

    vRefreshMCU ();

    // Falling edge detector for the external sync input

    if (oEXT_SYNCn == 0)
    {
      if (qPrevExtSyncSample == 1)
        qExtSyncEdge = TRUE;
      qPrevExtSyncSample = 0;
    }
    else
      qPrevExtSyncSample = 1;

    if (CCP1IE == 0)            // tick interrupt disabled = TX idling?
    {
      if (qExtSyncEdge || qGpsEvenMinute ()) // yes, is it time to start next frame now?
      {
        uns16 wTxProbability;

        // Generate a pseudo random number in the range 1..65536

        if (wRand == 0)         // zero is not allowed as it would lock wRand at 0 so...
          wRand = 0x2BAD;       // ...initialize once with any non-zero seed
        Carry = 0;
        wRand = rr (wRand);
        if (Carry)
        {
          wRand.high8 ^= 0xA1;
          wRand.low8 ^= 0xA1;
        }

        wTxProbability.low8 = bReadEE (eaTXPROB); // read probability threshold setting from EEPROM
        wTxProbability.high8 = bReadEE (eaTXPROB+1);

        if (wRand <= wTxProbability)          // did we get a "statistical yes"?
        {
          wTick10Length.low8 = bReadEE (eaBAUDCAL);     // get 16-bit signed baud calibration value from EEPROM
          wTick10Length.high8 = bReadEE (eaBAUDCAL+1);
          wTick10Length += 39322;                       // add nominal time for 10:th tick

          vInitDdsChip ();      // just in case the DDS was disconnected and reconnected...
          T1CON = 0b00000000;   // prescaler = 1, no oscillator, internal clock, TMR1 off
          CCP1CON = 0b00001010; // compare mode, software interrupt, CCP1 pin unaffected

          // Nominal start time for the first message symbol is top of even minute plus
          // 2 seconds. In case of GPS sync, we will delay the transmission start until
          // ~1.14s after receiving full even minute from the GPS. This allows for some
          // delay in the NMEA serial output but is still acceptable (WSPR sync search
          // is +/-5s according to Joe K1JT). In case of external pin sync, a start delay
          // of ~1.99s is used.

          TMR1L = 0;                          // start TMR1 from zero
          TMR1H = 0;
          CCPR1L = 0;                         // first compare interrupt after a full TMR1 cycle
          CCPR1H = 0;
          if (qExtSyncEdge)                   // external sync?
            jIsrGlobals.bTickCycleCount = 28; // yes, start delay = 28 full TMR1 cycles = ~1.99s
          else
            jIsrGlobals.bTickCycleCount = 16; // no, GPS, start delay = 16 full TMR1 cycles = ~1.14s
          jIsrGlobals.qPleaseQRT = FALSE;     // don't QRT prematurely...
          vEnqueueSymbol (bGetSymbol (0));    // load first symbol for transmission
          bSymbolIndex = 0;                   // symbol # currently in transmission
          CCP1IE = 1;                         // enable CCP1 interrupt (GIE & PEIE are set elsewhere)
          TMR1ON = 1;                         // GO!
        }

        qExtSyncEdge = FALSE;                 // we have responded to the sync flag, clear it

      } // end of "if any TX sync"
    }
    else      // CCP1IE == 1, ISR hasn't disabled itself yet, TX is still in progress
    if (jIsrGlobals.qNewFreqInQueue == FALSE)           // is th ISR ready for symbol update?
    {
      // When the transmission has been started, the ISR will run the
      // whole show in the background. Here, we just have to feed the ISR
      // with a new symbol as soon as it has consumed the previous one
      // (handshaking through the flag jIsrGlobals.qNewFreqInQueue).

      bSymbolIndex++;                                   // yes, bump symbol pointer
      if (bSymbolIndex < 162)                           // end of message?
        vEnqueueSymbol (bGetSymbol (bSymbolIndex));     // no, put the symbol in the queue
      else                                              // yes, transmission of last symbol has already started
      {
        vPrepareDdsForQsyTo (0);                        // make next symbol zero Hz = key-up
        jIsrGlobals.qPleaseQRT = TRUE;                  // tell ISR to key-up and quit after DDS update
        jIsrGlobals.qNewFreqInQueue = TRUE;
      }
    }

  } // end of main loop
}
